
class Sentence:
    """Represents a sentence or utterance."""

    def __init__(self, id_, words, acscore=None, lmscore=None):
        """Constructor.  ID and words are required."""
        self.id_ = id_
        self.words = words
        self.acscore = acscore
        self.lmscore = lmscore
        self.eval_ = None

    def __str__(self):
        """Returns a string representation of this object."""
        sentence_str = ' '.join(self.words).lower()
        return sentence_str

    def wer(self):
        if self.eval_:
            return self.eval_.wer()
        else:
            return 0.0

    def score(self, lmwt=10):
        return self.acscore + self.lmscore * lmwt



import logging

logger = logging.getLogger('asr_tools')
# logger.setLevel(logging.DEBUG)
logger.setLevel(logging.INFO)

# Create console handler and set level to debug
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
# Create formatter
# formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
formatter = logging.Formatter('%(levelname)s: %(message)s')
# Add formatter to ch
ch.setFormatter(formatter)
# Add ch to logger
logger.addHandler(ch)



import asr_tools.evaluation_util

from asr_tools.evaluation_util import evaluate
from asr_tools.evaluation_util import sum_evals
from asr_tools.evaluation_util import get_global_reference
from asr_tools.sentence_util import print_sentence


def nbest_oracle_sort(nbest, n=None):
    """Sort an n-best list by it's WER."""
    nbest = nbest.copy()
    if n: nbest.sentences = nbest.sentences[:n]
    nbest.sentences = sorted(nbest.sentences, key=lambda x: x.eval_.wer())
    return nbest

def nbest_best_sentence(nbest, n=None):
    """Find and return the sentence with the lowest WER in the n-best list."""
    sentences = nbest.sentences
    if n: sentences = sentences[:n]
    return min(sentences, key=lambda x: x.wer())

def nbest_oracle_eval(nbest, n=None):
    """Return the evaluation object of the best sentence in the nbest list."""
    return nbest_best_sentence(nbest, n=n).eval_

def evaluate_nbest(nbest, force=False):
    """Run our evaluation on each sentence in the nbest list.  Will skip evaluation if one
    is already there, unless forced to.  Saves the evaluation with each sentence."""
    id_ = nbest.id_
    for s in nbest.sentences:
        if force or s.eval_ is None:    # Only compute the evaluation if not already computed.
            e = evaluate(asr_tools.evaluation_util.REFERENCES, s)
            s.eval_ = e
    return nbest.sentences[0].eval_

def evaluate_nbests(nbests):
    """Return a single evaluation of list of n-best lists."""
    evals = list(map(evaluate_nbest, nbests))
    return sum_evals(evals)

def evaluate_nbests_oracle(nbests):
    """Return a single oracle evaluation for a list of n-best lists."""
    evals = list(map(nbest_oracle_eval, nbests))
    return sum_evals(evals)

def evals_by_depth(nbests, n=100):
    """Return overall oracle evaluations as a function of the depth of the n-best lists."""
    evals_by_depth = [None] * n
    for i in range(n):
        evals = []
        for nbest in nbests:
            evals.append(nbest_oracle_eval(nbest, i + 1))
        evals_by_depth[i] = sum_evals(evals)
    return evals_by_depth

def print_nbest(nbest, acscore=True, lmscore=True, tscore=True, tscore_wip=False,
                wcount=False, lmwt=10.0, maxwords=None, print_instances=False):
    """Returns a string representation of the object."""
    # This might be relatively slow because of all the string concatenation
    print_str = ''
    hyp = nbest.sentences[0]
    best = nbest_best_sentence(nbest)
    best_rank = nbest.sentences.index(best)
    print_str += 'ID: {} (#{} is best)\n'.format(nbest.id_, best_rank + 1)

    # Print reference if available
    ref = get_global_reference(nbest.id_)
    if ref:
        print_str += 'REF:  ' + str(ref) + '\n'
    else:
        print_str += '    No reference found.\n'
    print_str += 'HYP:  ' + str(hyp) + '\n'
    # print_str += 'BEST: '.format(best_rank + 1) + str(best) + '\n'
    print_str += 'BEST: ' + str(best) + '({})\n'.format(best_rank + 1)

    if print_instances:
        for i, s in enumerate(nbest.sentences):
            sentence_str = print_sentence(s, acscore=acscore, lmscore=lmscore, tscore=tscore,
                                          tscore_wip=tscore_wip, wcount=wcount, lmwt=lmwt, maxwords=maxwords)
            print_str += '{:3d} '.format(i + 1) + sentence_str
            if best_rank == i:
                print_str += ' **'
            print_str += '\n'
    print(print_str)

def print_nbest_ref_hyp_best(nbest):
    """Print three sentences: the reference, the top hypothesis, and the lowest WER
    hypothesis on the n-best list."""
    ref = get_global_reference(nbest.id_)
    hyp = nbest.sentences[0]
    best = nbest_best_sentence(nbest)
    best_rank = nbest.sentences.index(best)
    print_diff(ref, best, prefix1='REF: ', prefix2='BEST:')
    print_diff(best, hyp, prefix1='BEST:', prefix2='HYP: ')
    print('=' * 60)



import sys

from io import StringIO

from asr_tools.sentence import Sentence
from asr_tools.evaluation_util import get_global_reference
from asr_tools.evaluation_util import print_diff
from asr_tools.nbest_util import nbest_best_sentence

class NBest:
    """Represents an n-best list of ASR hypotheses."""

    sentences = None
    id_ = None

    def __init__(self, sentences, id_=None):
        """Sentences and IDs are required."""
        assert(sentences is not None)
        assert(len(sentences) > 0)
        self.sentences = sentences
        self.id_ = id_

    def __str__(self):
        """Returns a string representation of the object."""
        print_str = StringIO()
        print_str.write('ID: {}\n'.format(self.id_))
        for i, s in enumerate(self.sentences):
            print_str.write('{:3d} {}\n'.format(i + 1, s))
        return print_str.getvalue()

    def print_with_wer(self):
        """Returns a string representation of the object."""
        best = nbest_best_sentence(self)
        best_rank = self.sentences.index(best)
        print_str = StringIO()
        print_str.write('ID: {} (#{} is best)\n'.format(self.id_, best_rank))
        for i, s in enumerate(self.sentences):
            print_str.write('{:3d} {}'.format(i + 1, s))
            if best_rank == i:
                print_str.write(' **')
            print_str.write('\n')
        print(print_str.getvalue())

    def print_ref_hyp_best(self):
        """Print three sentences: the reference, the top hypothesis, and the lowest WER
        hypothesis on the n-best list."""
        ref = get_global_reference(self.id_)
        hyp = self.sentences[0]
        best = nbest_best_sentence(self)
        best_rank = self.sentences.index(best)
        print_diff(ref, best, prefix1='REF: ', prefix2='BEST:')
        print_diff(best, hyp, prefix1='BEST:', prefix2='HYP: ')
        print('=' * 60)

# This is another possible way to print the ref/hyp/best
# print_str = ''
# print_str += 'ID: {} (#{} is best)\n'.format(self.id_, best_rank)
# if get_global_reference(self.id_):
#     print_str += '{:3} '.format('') + str(ref) + '\n'
# else:
#     print_str += '    No reference found.\n'
# print_str += '{:3d} '.format(1) + str(hyp) + '\n'
# print_str += '{:3d} '.format(best_rank + 1) + str(best) + '\n'
# print(ref)



import logging

from collections import OrderedDict

import asr_tools.logging
from asr_tools.nbest import NBest
from asr_tools.sentence import Sentence

logger = logging.getLogger('asr_tools')


def read_transcript_table(f):
    """Given a file, read in the transcripts into a hash table
    indexed by IDs."""
    trans_table = OrderedDict()
    trans = read_transcript(f)
    for s in trans:
        trans_table[s.id_] = s
    return trans_table

def read_transcript(f):
    trans = []
    for line in f:
        line = line.strip()
        id_, words = line.split(maxsplit=1)
        s = Sentence(id_, words.split())
        trans.append(s)
    return trans

def read_nbest_file(f):
    "Read a Kaldi n-best file."
    nbests = []
    nbest = []
    current_id = None
    prev_id = None
    id_ = None
    while True:
        logger.debug('|NBESTS| = {}'.format(len(nbests)))
        logger.debug('|NBEST| = {}'.format(len(nbest)))
        entry = read_nbest_entry_lines(f)  # this is a sentence, which is spread across several lines
        logger.debug('ENTRY: ' + str(entry))
        if not entry:
            nbests.append(NBest(nbest, id_))
            logger.debug(nbest)
            break
            # yield NBest(nbest, id_)
        id_ = entry[0]
        id_ = id_.rsplit('-', maxsplit=1)[0]
        if not prev_id:
            prev_id = id_
        if id_ != prev_id:
            nbests.append(NBest(nbest, prev_id))
            nbest = []
            prev_id = id_
        s = entry_lines_to_sentence(entry)
        nbest.append(s)
    return nbests

def read_nbest_entry_lines(f):
    entry_lines = []
    while True:
        line = f.readline()
        if line == '':
            if len(entry_lines) == 0:
                return None
            else:
                return entry_lines
        if line == '\n':
            return entry_lines
        else:
            entry_lines.append(line.strip())

def entry_lines_to_sentence(lines):
    words = []
    lmscores = []
    acscores = []
    id_ = lines.pop(0)
    id_ = id_.rsplit('-', maxsplit=1)[0]
    for line in lines:
        tokens = line.split()
        if len(tokens) == 4:
            s1, s2, word, scores = tokens
            assert(int(s1) == int(s2) - 1)
            score_parts = scores.split(',')
            lmscores.append(float(score_parts[0]))
            acscores.append(float(score_parts[1]))
            words.append(tokens[2])
    lmscore = sum(lmscores)
    acscore = sum(acscores)
    return Sentence(id_, words, lmscore=lmscore, acscore=acscore)

# def read_nbest_file(f):
#     "Read a Kaldi n-best file."
#     nbest = []
#     current_id = None
#     prev_id = None
#     id_ = None
#     while True:
#         entry = read_nbest_entry_lines(f)  # this is a sentence.
#         # logger.info('TEST')
#         # print(entry)
#         if not entry:
#             # print(nbest)
#             break
#             # yield NBest(nbest, id_)
#         id_ = entry[0]
#         id_ = id_.rsplit('-', maxsplit=1)[0]
#         if not prev_id:
#             prev_id = id_
#         if id_ != prev_id:
#             yield NBest(nbest, prev_id)
#             nbest = []
#             prev_id = id_
#         s = entry_lines_to_sentence(entry)
#         nbest.append(s)



import unittest

from asr_tools.kaldi import read_nbest_file
from asr_tools.kaldi import read_transcript_table
from asr_tools.kaldi import read_transcript
from asr_tools.evaluation_util import evaluate

class Testing(unittest.TestCase):

    nbest_file = "test/data/librispeech1/lat.1.nbest.txt"
    ref_file = "test/data/librispeech1/dev_clean.ref"
    hyp_file = "test/data/librispeech1/dev_clean.14_0.5.tra.txt"

    # Make sure we can read the main (3) types of files.
    def test_read_nbest(self):
        with open(self.nbest_file) as f:
            nbests = list(read_nbest_file(f))
            self.assertTrue(len(nbests) == 40)

    def test_read_transcript(self):
        with open(self.ref_file) as f:
            refs = read_transcript_table(f)
            self.assertTrue(len(refs) == 2703)

    def test_read_hyp(self):
        with open(self.hyp_file) as f:
            refs = read_transcript(f)
            self.assertTrue(len(refs) == 2703)

    # This compares a hyp file to a ref file--this test is probably too slow
    def test_evaluation(self):
        with open(self.hyp_file) as h, open(self.ref_file) as r:
            ref_table = read_transcript_table(r)
            hyps = read_transcript(h)
            evals = []
            for hyp in hyps:
                ref = ref_table[hyp.id_]
                eval_ = evaluate(ref_table, hyp)
                evals.append(eval_)
            overall_eval = sum(evals[1:], evals[0])
            self.assertTrue(overall_eval.ref_len == 54402)
            self.assertTrue(overall_eval.matches == 51159)
            self.assertTrue(overall_eval.errs == 3815)

    # python ./bin/sklearn-test.py  ~/data/librispeech1/nbests/lat.1.nbest.txt ~/data/librispeech1/dev_clean.ref
    # python ./bin/test-fe.py  ~/data/librispeech1/nbests/lat.1.nbest.txt
    # Other things to test:
    # Computing Oracle WER?
    # python ./bin/perceptron-training.py  ~/data/librispeech1/nbests/lat.1.nbest.txt.testing.2x \
    #    ~/data/librispeech1/dev_clean.ref



from collections import OrderedDict

from asr_tools.evaluation import Evaluation
from edit_distance import SequenceMatcher, edit_distance
from asr_evaluation.asr_evaluation import print_diff as eval_print_diff

REFERENCES = OrderedDict()

def evaluate_hyps(hyps, ref_table):
    evals = []
    for hyp in hyps:
        ref = ref_table[hyp.id_]
        eval_ = evaluate(ref_table, hyp)
        evals.append(eval_)
    return sum(evals[1:], evals[0])

def sentence_editdistance(s1, s2):
    """Given two 'sentence' objects compute the edit distance and
    return the distance."""
    distance, matches = edit_distance(s1.words, s2.words)
    return distance

def evaluate(ref_table, s):
    """Given a sentence and a reference table, create and return an
    Evaluation object."""
    ref = ref_table.get(s.id_)
    if ref is None:
        raise Exception('No reference loaded for ID: {}'.format(s.id_))
    distance, matches = edit_distance(ref.words, s.words)
    eval_ = Evaluation(len(ref.words), matches, distance)
    return eval_

def set_global_references(ref_file):
    """Given a reference file, read it into the global table with the name
    REFERENCES."""
    global REFERENCES
    REFERENCES = read_transcript_table(ref_file)

def get_global_reference(id_):
    """Look-up ASR references by id."""
    global REFERENCES
    return REFERENCES.get(id_)

def print_diff(s1, s2, prefix1='REF:', prefix2='HYP:'):
    """Print a readable diff between two sentences."""
    a = s1.words
    b = s2.words
    sm = SequenceMatcher(a, b)
    eval_print_diff(sm, s1.words, s2.words, prefix1=prefix1, prefix2=prefix2)

def sum_evals(evals):
    if len(evals) == 1:
        return evals[0]
    else:
        return sum(evals[1:], evals[0])



# In my previous version, I had a couple versions of evaluation object.
# Some contained the words, etc. that caused errors.
# We can use this same object for sentence error rate.

class Evaluation():
    ref_len = None
    matches = None
    errs = None

    def __init__(self, ref_len, matches, errs):
        self.ref_len = ref_len
        self.matches = matches
        self.errs = errs

    def wer(self):
        """Calculate word error rate (WER)"""
        return self.errs / self.ref_len

    def wrr(self):
        """Calculate word recognition rate (WRR)"""
        return self.matches / self.ref_len

    def is_correct():
        return errs == 0

    def __add__(self, other):
        """Add each field, creating a new object."""
        return Evaluation(self.ref_len + other.ref_len,
                          self.matches + other.matches,
                          self.errs + other.errs)

    def __str__(self):
        """Print the primary evaluation metrics."""
        lines = ["Ref length: {:>10,d}".format(self.ref_len),
                 "Matches:    {:>10,d}".format(self.matches),
                 "Errors:     {:>10,d}".format(self.errs),
                 "WER:        {:>10.2%}".format(self.wer())]
        return '\n'.join(lines)

    def __cmp__(self, other):
        self.wer() - other.wer()



from setuptools import setup

setup(
    name='asr_tools',
    version='0.1',
    author='Ben Lambert',
    author_email='ben@benjaminlambert.com',
    packages=['asr_tools'],
    description='Automatic speech recognition tools.',
    url='https://github.com/belambert/asr-tools',
    keywords=['automatic speech recognition', 'speech recognition', 'asr'],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering",
        "License :: OSI Approved :: BSD License"],
    # install_requires=['edit_distance', 'asr_evaluation'],
    test_suite='test.test_main.Testing'
)



from io import StringIO

def print_sentence(s, acscore=True, lmscore=True, tscore=True, tscore_wip=False,
                   wcount=False, lmwt=10.0, maxwords=None, wer=True):
    """Prints all sentence scores including acoustic, lm, overall, etc."""
    output = StringIO()
    number_template = '{:8,.2f}'
    spaces = 2
    if maxwords:
        sentence_str = ' '.join(s.words[:maxwords]).lower()
    else:
        sentence_str = ' '.join(s.words).lower()
    if acscore:
        output.write(number_template.format(s.acscore) + ' ' * spaces)
    if lmscore:
        output.write(number_template.format(s.lmscore) + ' ' * spaces)
    if tscore:
        output.write(number_template.format(s.score(lmwt=lmwt)) + ' ' * spaces)
    if tscore_wip:
        output.write(number_template.format(s.score(lmwt=lmwt) + len(s.words) * 0.5) + ' ' * spaces)
    if wcount:
        output.write('{:5}'.format(len(s.words)) + ' ' * spaces)
    if wer and s.eval_:
        output.write('{:5.0%}  '.format(s.eval_.wer()))

    output.write(sentence_str)
    return output.getvalue()





def rerank_nbests(nbests, func):
    """Given a a list of nbests and a function, rerank all the nbests
    with the given function."""
    for nbest in nbests:
        rerank_nbest(nbest, func)

def rerank_nbest(nbest, func):
    """Given an n-best list and a function, rerank the n-best list
    using that function."""
    nbest.sentences = sorted(nbest.sentences, key=func)
